#!/usr/bin/perl
#
# zf-Webmail
# zeitform Internet Dienste <zeitform@zeitform.de>
# http://www.zeitform.de
# Version 0.6, Aug 2001
#
# Zu Installation und Konfiguration siehe die Datei README
# Dieses Skript und alle begleitenden Dateien unterliegen der GNU General 
# Public Licence (GPL) (siehe http://www.fsf.org/copyleft/gpl.html).

### Konfiguration
my ($config_dir) = $ENV{SCRIPT_FILENAME} =~ /^(.+)\//;
require "$config_dir/webmail_config.pm" or die "cannot read configuration";

### Perl Module
use strict;
use CGI;
use HTML::Template;
use Net::POP3;
use Net::SMTP;
use Crypt::CBC;
use MIME::Parser;
use Date::Parse;
#use CGI::Carp 'fatalsToBrowser';

use vars qw($entity);

## Variablen zum Aufbewahren der Mails
my (@body, @attachment, @bodypath);

## CGI-Handler und Aufruf-Typ zur weiteren Verzweigung
my $q = new CGI;
my $type = $q->param("type");

# Die Variable $type enthlt eine vitale Information fr
# dieses Skript. Ist sie nicht definiert, so erhlt der
# Benutzer das Anmeldeformular. $type legt fest, welche
# Aktion ausgefhrt bzw. welche Seite angezeigt werden soll.

############################################################
if ($type eq "") { # zeige Anmeldeseite
############################################################

  my $template = HTML::Template->new(filename => $Conf::login_tmpl);

  $template->param(
		   CGI => $ENV{SCRIPT_NAME},
		   SESSION => $ENV{UNIQUE_ID},
		  );

  # lsche vorhandene Cookies und belege sie mit
  # "harmlosen Werten"
  my $cookie = $q->cookie(-name => "zf_webmail",
			  -value => "zeitform",
			  -expires => '+10m');

  print $q->header(-cookie=>$cookie, -expires => "now");
  print $template->output;

############################################################
} elsif ($type eq "login") { # Benutzer ist angemeldet
############################################################

  # die relevanten Informationen werden aus dem POST ausgelesen
  my $login    = $q->param("login")    or print_error("no_login");
  my $password = $q->param("password") or print_error("no_password");
  my $pophost  = $q->param("pop3")     or print_error("no_pop3");
  my $smtphost = $q->param("smtp")     or print_error("no_smtp");

  # falls bestimmte SMTP-Server ausgeschlossen werden mssen
  if ($Conf::exclude_some_smtp_hosts) {
    foreach (namelookup($smtphost)) {
      print_error('smtp_excluded') if $Conf::smtp_exclude_list{$_};
    }
  }

  my $session = $q->param("session") or print_error("no_session");
  my $key;

  # und an die Mail-bersicht weitergegeben
  show_folder($login, $password, $pophost, $smtphost, $session);

############################################################
} elsif ($type eq "view") { # zeige eine Email
############################################################

  # Cookie auslesen
  my ($login, $password, $pophost, $smtphost, $session) = getcookie();

  # Nachrichten-Nummer auslesen
  my $msg = $q->param("msg") or print_error("no_msg");

  # kontaktiere den POP3-Server und hole den Mail-Body
  my $pop = Net::POP3->new($pophost, Timeout => 60);
  print_error("login_failed") unless ($pop);

  my $ok = $pop->apop($login, $password);

  unless ($ok) {### apop nicht moeglich, daher versuchen wir pop3
      $pop->reset();
      $pop->quit();
      $pop = Net::POP3->new($pophost, Timeout => 60);
      print_error("login_failed") unless ($pop);
      $ok = $pop->login($login, $password);
      print_error("login_failed") unless $ok;

  }


  print_error("no_mails") if $ok == 0;

  my $message = $pop->get($msg);


  # parse die Email auf MIME-Anhnge
  my $parser = new MIME::Parser;
  $parser->output_under($Conf::msg_dir);
  my $entity = $parser->parse_data($message);

  split_entity($entity);

  # Variablen fr die Links "vorherige" und "nchste"
  my $next_msg = $msg - 1;
  my $prev_msg = $msg + 1;
  my $next_ok  = $next_msg > 0 ? 1 : 0;
  my $prev_ok  = $prev_msg <= $ok ? 1 : 0;

  # flle die Template
  my $template = HTML::Template->new(filename => $Conf::mail_tmpl);
  $template->param(
		   from    => CGI::escapeHTML($entity->head->decode->get("From")),
		   subject => CGI::escapeHTML($entity->head->decode->get("Subject")),
		   date    => stringify_date($entity->head->decode->get("Date")),
		   to      => CGI::escapeHTML($entity->head->decode->get("To")),
                   body    => [ @body ],
		   attach  => [ @attachment ],
                   DELETE_LINK => $ENV{SCRIPT_NAME} . "?type=delete&msg=" . 
                              CGI::escape($msg . "::" . $pop->uidl($msg)),
		   REPLY_LINK  => $ENV{SCRIPT_NAME} . "?type=reply&msg=$msg",
		   FWD_LINK => $ENV{SCRIPT_NAME} . "?type=fwd&msg=$msg",
		   COMPOSE_LINK => $ENV{SCRIPT_NAME} . "?type=compose",
		   NEXT    => $ENV{SCRIPT_NAME} . "?type=view&msg=" . $next_msg,
		   PREV    => $ENV{SCRIPT_NAME} . "?type=view&msg=" . $prev_msg,
                   next_ok => $next_ok,
		   prev_ok => $prev_ok,
		   FOLDER  => $ENV{SCRIPT_NAME} . "?type=folder",
		   CGI     => $ENV{SCRIPT_NAME},
		  );

  $pop->reset(); 
  $pop->quit();

  my $cookie = setcookie($login, $password, $pophost, $smtphost, $session);

  print $q->header(-cookie=>$cookie, -expires => "now");
  print $template->output;

############################################################
} elsif ($type eq "compose") { # schreibe eine neue Mail
############################################################

  # Cookie auslesen
  my ($login, $password, $pophost, $smtphost, $session) = getcookie();

  # flle Template
  my $template = HTML::Template->new(filename => $Conf::reply_tmpl);
  $template->param(
		   to      => "",
		   from    => $login,
		   subject => "",
		   body    => "",
                   msg     => "",
		   CGI     => $ENV{SCRIPT_NAME},
                   fwd     => "",
                   attach  => "",
                   attachurl => "",
		  );

  my $cookie = setcookie($login, $password, $pophost, $smtphost, $session);

  print $q->header(-cookie=>$cookie, -expires => "now");
  print $template->output;


############################################################
} elsif ($type eq "reply" || $type eq "fwd") { # schreibe oder beantworte eine Mail
############################################################

  # Cookie auslesen
  my ($login, $password, $pophost, $smtphost, $session, $key) = getcookie();

  # Nachrichten-Nummer auslesen
  my $msg = $q->param("msg") or print_error("no_msg");

  # kontaktiere den POP3-Server und hole den Mail-Body
  my $pop = Net::POP3->new($pophost, Timeout => 60);
  print_error("login_failed") unless ($pop);

  my $ok = $pop->apop($login, $password);

  unless ($ok) {### apop nicht moeglich, daher versuchen wir pop3
      $pop->reset();
      $pop->quit();
      $pop = Net::POP3->new($pophost, Timeout => 60);
      print_error("login_failed") unless ($pop);
      $ok = $pop->login($login, $password);
      print_error("login_failed") unless $ok;
  
  }

  print_error("no_mails") if $ok == 0;

  my $message = $pop->get($msg);


  # parse die Email auf MIME-Anhnge
  my $parser = new MIME::Parser;
  $parser->output_under($Conf::msg_dir);
  my $entity = $parser->parse_data($message);

  split_entity($entity);

  my $date    = $entity->head->decode->get("Date");    $date    =~ s/\n//g;
  my $to      = $entity->head->decode->get("To");      $to      =~ s/\n//g;
  my $from    = $entity->head->decode->get("From");    $from    =~ s/\n//g;
  my $subject = $entity->head->decode->get("Subject"); $subject =~ s/\n//g;
  my $body    = ""; 
  my $attach; 
  my $attachurl;
  my $quote;    # for quoting

  if ($type eq "reply") { # reply

    $subject = "Re: " .$subject if $subject;
    $quote = "> ";
    $to = $from;
    $from = $login;


  } else { # forward

    foreach (@attachment) {
      $attachurl .= $_->{url} . ":" . $_->{type} . "::";
      $attach    .= $_->{name} . ", ";
    }

    $attach =~ s/, $//;

    $body = "From: $from\nTo: $to\nDate: $date\nSubject: $subject\n";
    $body .= "Attachments: $attach\n" if $attach;
    $body .= "\n";
    $subject = "Fwd: " .$subject if $subject;
    $from = $login; 
    $to = "";

  }

  # lese Mailtext fr reply
  foreach (@bodypath) {
    open IN, "$Conf::msg_dir$_" or next;
    while (<IN>) {
      $body .= $quote . $_;
    }
    close IN;
  }

  # flle Template
  my $template = HTML::Template->new(filename => $Conf::reply_tmpl);
  $template->param(
		   to        => CGI::escapeHTML($to),
		   from      => CGI::escapeHTML($from),
		   subject   => CGI::escapeHTML($subject),
		   body      => CGI::escapeHTML($body),
                   msg       => $msg,
		   CGI       => $ENV{SCRIPT_NAME},
                   fwd       => ($type eq "fwd" && $attach),
                   attach    => $attach,
                   attachurl => $attachurl,
		  );


  $pop->reset(); 
  $pop->quit();

  my $cookie = setcookie($login, $password, $pophost, $smtphost, $session);

  print $q->header(-cookie=>$cookie, -expires => "now");
  print $template->output;


############################################################
} elsif ($type eq "send_reply") { # sende die Mail
############################################################

  # lese Cookie
  my ($login, $password, $pophost, $smtphost, $session, $key) = getcookie();

  # lese Parameter
  my $msg          = $q->param("msg");
  my $to           = $q->param("to")      or print_error("no_recipient");
  my $cc           = $q->param("cc");
  my $from         = $q->param("from")    or print_error("no_sender");
  my $subject      = $q->param("subject");
  my $text         = $q->param("text");   $text =~ s/\r//g; # needed!
  my $oldattachurl = $q->param("attach0url");
  my $oldattach    = $q->param("attach0");

  # manage date
  # maybe switch to Date::Format here ...
  my @weekday = (qw(Sun Mon Tue Wed Thu Fri Sat));
  my @month   = (qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec));
  my @today = localtime; 
  $today[5] += 1900; 
  $today[4] = $month[$today[4]]; 
  $today[6] = $weekday[$today[6]];
  $today[8] += $Conf::timezone;
  my $today = sprintf("%s, %d %s %d %02d:%02d:%02d %+05d", @today[6,3,4,5,2,1,0,8]);
 
  # zhle Anhnge
  my $i = 1; my @attach;
  while ($q->param("attach$i")) {
    push @attach, $q->param("attach$i");
    $i++;
  }

  # generiere neue Mail
  # problematisch ist hier das Setzen des korrekten Content-Type
  # Bitte melden Sie sich, wenn es hier zu Problemen kommt.
  my $mail = MIME::Entity->build(From     => $from,
				 To       => $to,
				 Subject  => $subject,
				 Encoding => "quoted-printable",
				 "Content-Type" => 'text/plain; charset="iso-8859-1"', ## ???
				 Cc       => $cc,
				 Date     => $today,
                                 @Conf::xheader,
				 Data     => $text);

  # hnge alte Anlagen an (nur bei forward)
  if ($oldattach) {
    foreach (split(/,\s*/, $oldattach)) {
      if ($oldattachurl =~ /([^:]+\/$_):([^:]+)/) {
	if (-r "$ENV{DOCUMENT_ROOT}/$1" ) {
	  $mail->attach(Path => "$ENV{DOCUMENT_ROOT}/$1", 
			Type => "$2",
			Encoding => "base64");
	}
      }
    }
  }

  # hnge Anlagen an
  foreach (@attach) {
    my $buffer;
    my $bytesread;
    open (OUTFILE,">$Conf::tmp_dir/$$-$_") or print_error("attachment_failed");
    while ($bytesread=read($_,$buffer,1024)) {
      print OUTFILE $buffer;
    }
    close OUTFILE;

    $mail->attach(Path     => "$Conf::tmp_dir/$$-$_", 
                  Type     => $q->uploadInfo($_)->{'Content-Type'},
		  Encoding => "base64");
  }

  # kommuniziere mit dem SMTP-Server und sende die Mail

  # Es wre vorteilhaft, hier bei einer Fehlermeldung, die auf einen
  # SMTP-Server hinweist, der kein Relaying untersttzt (zumindest
  # nicht in der Form, in der wir es hier bentigen), dem Benutzer
  # die Mglichkeit zu geben, einen anderen SMTP-Server zu whlen,
  # ohne ihn seine Email erneut eingeben zu lassen.

  my $smtp = Net::SMTP->new($smtphost, Timeout => 60, Hello => $smtphost) 
    or print_error("smtp_unreachable");

  $smtp->mail($from)                                or print_error("smtp_login_failed");
  $smtp->to(split(/,\s*/, $to), split(/,\s*/, $cc)) or print_error("no_relay");
  $smtp->data($mail->as_string)                     or print_error("smtp_failed");
  $smtp->quit;

  # flle Template
  my $template = HTML::Template->new(filename => $Conf::reply_success_tmpl);
  $template->param(
		   to      => CGI::escapeHTML($to),
		   from    => CGI::escapeHTML($from),
		   subject => CGI::escapeHTML($subject),
		   cc      => CGI::escapeHTML($cc),
		   VIEW_LINK => ($msg ? $ENV{SCRIPT_NAME} . "?type=view&msg=$msg" :
		                        $ENV{SCRIPT_NAME} . "?type=folder"),
		  );

  my $cookie = setcookie($login, $password, $pophost, $smtphost, $session);

  print $q->header(-cookie=>$cookie, -expires => "now");
  print $template->output;

############################################################
} elsif ($type eq "delete") { # lsche Mail(s)
############################################################

  # lese Nachrichten-Nummer und Cookie
  my @msg = $q->param("msg");
  my ($login, $password, $pophost, $smtphost, $session) = getcookie();

  # Kontaktiere POP3-Server
  my $pop = Net::POP3->new($pophost, Timeout => 60);
  print_error("login_failed") unless ($pop);

  my $ok = $pop->apop($login, $password);

  unless ($ok) {### apop nicht moeglich, daher versuchen wir pop3
      $pop->reset();
      $pop->quit();
      $pop = Net::POP3->new($pophost, Timeout => 60);
      print_error("login_failed") unless ($pop);
      $ok = $pop->login($login, $password);
      print_error("login_failed") unless $ok;

  }

  print_error("no_mails") if $ok == 0; # kann eigentlich nicht passieren

  # lsche die Nachrichten
  foreach (@msg) {
    my ($msg, $uidl) = split /::/, $_;
    if ($pop->uidl($msg) eq $uidl) {
      $pop->delete($_);
    }
  }

  # beende die Verbindung. Kein $pop->reset() hier, das die 
  # Lschungen sonst nicht durchgefhrt werden wrden.
  $pop->quit();

  # zeige bersicht
  show_folder($login, $password, $pophost, $smtphost, $session);


############################################################
} elsif ($type eq "folder") { # zeige bersicht
############################################################

  # zeige bersicht
  my ($login, $password, $pophost, $smtphost, $session) = getcookie();

  show_folder($login, $password, $pophost, $smtphost, $session);

############################################################
} else { # es gibt keine weiteren Werte fr $type
############################################################

  print_error("illegal_call");

############################################################
}
############################################################




### subs


############################################################
sub print_error {
############################################################

  my $error_key = shift || "default";
  my $error = $Conf::error_msg{$error_key} || $Conf::error_msg{default};

  my $template = HTML::Template->new(filename => $Conf::error_tmpl);

  $template->param( ERROR => $error );

  print $q->header(-expires => "now");
  print $template->output;
  exit 1;

}

############################################################
sub show_folder {
############################################################

  my ($login, $password, $pophost, $smtphost, $session) = @_;
  # check if present
  print_error("login_failed") unless ($login && $password && $pophost);

  my $pop = Net::POP3->new($pophost, Timeout => 60);
  print_error("login_failed") unless ($pop);

  # eine bessere Fehlerbehandlung wrde dem Kunden hier keinen Fehler
  # zurckliefern, wenn er keine Emails auf dem Server leigen hat,
  # sondern lediglich die entsprechende Nachricht.

  my $ok = $pop->apop($login, $password);

  unless ($ok) {### apop nicht moeglich, daher versuchen wir pop3
      $pop->reset();
      $pop->quit();
      $pop = Net::POP3->new($pophost, Timeout => 60);
      print_error("login_failed") unless ($pop);
      $ok = $pop->login($login, $password);
      print_error("login_failed") unless $ok;
  }

  ###print_error("no_mails") if $ok == 0; ### hier muessen wir was anderes anzeigen!!!

  if ($ok == 0) { ### keine mails vorhanden. 


    my $template = HTML::Template->new(filename => $Conf::empty_folder_tmpl);

    $template->param(
		     CGI          => $ENV{SCRIPT_NAME},
		     COMPOSE_LINK => $ENV{SCRIPT_NAME} . "?type=compose",
		     RELOAD_LINK  => $ENV{SCRIPT_NAME} . "?type=folder",
		    );

    my $cookie = setcookie($login, $password, $pophost, $smtphost, $session);

    print $q->header(-cookie=>$cookie, -expires => "now");
    print $template->output;

  } else { ### mails sind vorhanden und mssen angezeigt werden

    my $list = $pop->list();
    print_error("list_failed") unless $list; #????

    my $template = HTML::Template->new(filename => $Conf::folder_tmpl);

    my $msg;
    my @loop;
    foreach $msg (sort { $b <=> $a }keys %$list) {
      my $head = new MIME::Head $pop->top($msg); ## Mail::Header

      push @loop, { from      => CGI::escapeHTML(shorten($head->decode->get("From"), $Conf::max_from))
                                  || "[unknown]",
		    subject   => CGI::escapeHTML(shorten($head->decode->get("Subject"), $Conf::max_subject))
                                  || "[no subject]",
		    date      => stringify_date(CGI::escapeHTML($head->decode->get("Date"))),
		    size      => int($list->{$msg}/1024+0.5) . " K",
		    msg       => CGI::escapeHTML($msg . "::" . $pop->uidl($msg)),
		    VIEW_LINK => $ENV{SCRIPT_NAME} . "?type=view&msg=" . $msg,
		  };
    }

    $pop->reset();
    $pop->quit();

    $template->param(
		     MAILS        => [ @loop ],
		     DELETE_LINK  => $ENV{SCRIPT_NAME},
		     CGI          => $ENV{SCRIPT_NAME},
		     NUM          => $ok,
                     RELOAD_LINK  => $ENV{SCRIPT_NAME} . "?type=folder",
		     COMPOSE_LINK => $ENV{SCRIPT_NAME} . "?type=compose",
		    );

    my $cookie = setcookie($login, $password, $pophost, $smtphost, $session);

    print $q->header(-cookie=>$cookie, -expires => "now");
    print $template->output;

  }
}


############################################################
sub setcookie {
############################################################

  my $login    = shift;
  my $password = shift;
  my $pophost  = shift;
  my $smtphost = shift;

  my $session  = shift;

  my $key = $ENV{UNIQUE_ID};

  my $ciphertext;

  if ($Conf::encryption_method) {
    my $cipher = new Crypt::CBC($key, $Conf::encryption_method);
    $ciphertext = $cipher->encrypt($password);

    open ID, ">$Conf::tmp_dir/$session" or print_error("write_error");
    print ID $key;
    close ID;
  } else { # keine Verschlsselung
    $ciphertext = $password;
  }

  my $cookie = $q->cookie(-name    => "zf_webmail",
			  -value   => {
				       pop3     => $pophost,
				       smtp     => $smtphost,
				       login    => $login,
				       password => $ciphertext,
				       id       => $session,
				      },
			  -expires => '+10m');

  return $cookie;

}

############################################################
sub getcookie {
############################################################

  my %cookies = $q->cookie(-name => "zf_webmail");
  my $login = $cookies{login};

  unless ($login) { # kein cookie, keine session!!!

    my $template = HTML::Template->new(filename => $Conf::expired_tmpl);

    $template->param(
		     CGI     => $ENV{SCRIPT_NAME},
		     SESSION => $ENV{UNIQUE_ID},
		    );

    print $q->header(-expires => "now");
    print $template->output;
    exit 0;

  } else {

    my $ciphertext = $cookies{password} or print_error("no_cook_password");
    my $pophost    = $cookies{pop3}     or print_error("no_cook_pop3");
    my $smtphost   = $cookies{smtp}     or print_error("no_cook_smtp");
    my $session    = $cookies{id}       or print_error("no_cook_session");

    my $password;

    if ($Conf::encryption_method) {
      open ID, "$Conf::tmp_dir/$session" or print_error("read_error");
      my $key = <ID>;
      close ID;

      my $cipher = new Crypt::CBC($key, $Conf::encryption_method);
      $password = $cipher->decrypt($ciphertext);
    } else { # keine Verschlsselung
      $password = $ciphertext;
    }

    return ($login, $password, $pophost, $smtphost, $session);
  }

}


############################################################
sub split_entity {
############################################################
  local $entity = shift;
  my $num_parts = $entity->parts;      # how many mime parts?

  if ($num_parts) {                    # we have a multipart mime message

    foreach (1..$num_parts) {
      split_entity( $entity->parts($_ - 1) );
    }

  } else {                             # we have a single mime message/part

    # we should handle the next line different (config?)
    if ($entity->effective_type =~ /^text\/(?!(html|enriched))/) {
      my $path = $entity->bodyhandle->path;
      $path =~ s/$Conf::msg_dir//;
      push @bodypath, $path;
      push @body, { text => beautify(CGI::escapeHTML($entity->bodyhandle->as_string)) };
    } else {
      my $path = $entity->bodyhandle->path;
      $path =~ s/$Conf::msg_dir/$Conf::attach_url/;
      # we must escape this, but not the "/"
      $path = CGI::escape($path); $path =~ s/%2F/\//g;
      my ($cleanpath) = ($path =~ /([^\/]+)$/);
      push @attachment, { url => $path, name => $cleanpath, type => $entity->effective_type };
    }
  }
}

############################################################
sub shorten {
############################################################

  my $string = shift || return "";
  my $size =   shift || return "";

  if (length($string) <= $size) {
    return $string;
  } else {
    return substr($string, 0, $size) . "...";
  }
}

############################################################
sub stringify_date {
############################################################

  my $date = shift || return "";

  my ($ss,$mm,$hh,$day,$month,$year,$zone) = strptime($date);

  return "" unless $day && $hh;
  return sprintf("%02d.%02d.%04d %02d:%02d [%+dh]", 
		 $day, $month+1, $year+1900, $hh, $mm, $zone/3600);
}


############################################################
sub beautify {
############################################################
  # we mainly enhance the body text here
  my $text = shift;

  $text =~ s/(http:\/\/\S+)/<a href=\"$1\">$1<\/a>/g;
  $text =~ s/\n/<br>\n/g;
  $text =~ s/\s\s/ &nbsp;/g;
  #$text =~ s/Microsoft/Micro\$oft/g; # This is just a joke!

  return $text;
}

############################################################
sub namelookup {
############################################################
  my $host = shift;
  my @result;

  my ($name, $alias, $addrtyp, $length, @addr) = (gethostbyname $host);

  foreach (@addr) {
    push @result, join(".", unpack('C4', $_));
  }

  return @result;
}



###  fin
